# -*-coding:utf-8-*-

__author__ = 'karlvorndoenitz@gmail.com'


RESPONSE_HEAD_MESSAGE = "HTTP/1.1 %(status)s %(status_msg)s\r\n" \
                        "Date: %(date)s\r\n" \
                        "Host: %(host)s\r\n" \
                        "Content-Length: %(content_length)s\r\n\r\n"
SOCKET_RECEIVE_SIZE = 1024 * 64
DEFAULT_PORT = 80
COOKIE_SECURITY_DEFAULT_STRING = "1qaz2wsx3"
CLIENT_CONNECT_TO_SERVER_NUM = 128

content_type = {
    "jpg": "image/jpeg",
    "jpeg": "image/jpeg",
    "png": "image/png",
    "gif": "image/gif",
    "css": "text/css",
    "js": "text/js",
    "mp3": "audio/mpeg",
    "ogg": "audio/ogg",
    "mp4": "video/mp4"
}


class HttpStatus(object):
    SUCCESS = 200
    PARTIAL_SUCCESS = 206
    REDIRECT = 302
    RESOURCE_NOT_MODIFIED = 304
    REQUEST_FORBID = 403
    NOT_FOUND = 404
    METHOD_NOT_ALLOWED = 405
    SERVER_ERROR = 500
    SERVER_NOT_SUPPORT = 501
    TIME_OUT = 504


class HttpStatusMsg(object):
    SUCCESS = "OK"
    PARTIAL_SUCCESS = "OK"
    REDIRECT = "Moved Temporarily"
    RESOURCE_NOT_MODIFIED = "Not Modified"
    REQUEST_FORBID = "FORBID"
    NOT_FOUND = "NOT-FOUND"
    METHOD_NOT_ALLOWED = "METHOD-NOT-ALLOWED"
    SERVER_ERROR = "SERVER-ERROR"
    SERVER_NOT_SUPPORT = "NOT-SUPPORT"
    TIME_OUT = "TIME-OUT"


class ContentType(object):
    FORM = "application/x-www-form-urlencoded"
    JSON = "application/json"
    HTML = "text/html"
    TEXT = "text/plain"



# -*-encoding:utf-8-*-


class HttpConnection(object):
    """

    This class contains four methods

    def __init__(self): initialize

    def add_connection(self, name, value): add connection

    def get_connection(self, name): get connection

    def remove_connection(self, name): remove connection

    """
    def __init__(self):
        self.connection = {}
        self.requests = {}
        self.responses = {}

    def add_connection(self, name, value):
        """add connection

        :param name: connection name
        :param value: connection
        :return: None

        """
        self.connection[name] = value

    def get_connection(self, name):
        """get connection

        :param name: connection name
        :return: connection

        """
        return self.connection[name]

    def remove_connection(self, name):
        """remove connection

        :param name: connection name
        :return: None

        """
        if name in self.connection:
            del self.connection[name]



# -*-coding:utf-8-*-

from jinja2 import Template
from karlooper.utils import PY3
import sys

if not PY3:
    reload(sys)
    sys.setdefaultencoding('utf-8')

__author__ = 'karlvorndoenitz@gmail.com'


def render(template_path, **kwargs):
    """read template from file and compile to html content

    :param template_path: template file's path
    :param kwargs: parameters
    :return: html content

    """
    with open(template_path) as template_file:
        template_data = template_file.read()
    template = Template(template_data)
    data = template.render(kwargs)
    return data


def render_string(template_string, **kwargs):
    """compile template string to html content

    :param template_string: template string
    :param kwargs: parameters
    :return: html content

    """
    template = Template(template_string)
    return template.render(kwargs)



# -*-encoding:utf-8-*-


class HttpIOBuffer(object):
    """

    This class contains seven methods:

    def add_request(self, name, value): add request to io buffer

    def add_response(self, name, value): add response to io buffer

    def get_request(self, name): get request from io buffer

    def get_response(self, name): get response from io buffer

    def remove_request(self, name): get request from io buffer

    def remove_response(self, name): get response from io buffer

    """
    def __init__(self):
        self.requests = {}
        self.responses = {}

    def add_request(self, name, value):
        """add request

        :param name: request's name
        :param value: request
        :return: None

        """
        self.requests[name] = value

    def add_response(self, name, value):
        """add response

        :param name: response's name
        :param value: response
        :return: None

        """
        self.responses[name] = value

    def get_request(self, name):
        """get request

        :param name: the name of request
        :return: request

        """
        return self.requests[name]

    def get_response(self, name):
        """get response

        :param name: the name of response
        :return: response

        """
        return self.responses[name]

    def remove_request(self, name):
        """remove request

        :param name: the name of request
        :return: None

        """
        if name in self.requests:
            del self.requests[name]

    def remove_response(self, name):
        """remove response

        :param name: the name of response
        :return: None

        """
        if name in self.responses:
            del self.responses[name]



# -*-coding:utf-8-*-
"""

__async_core_server
~~~~~~~~~~~~~~~~~~

introduction
use python asyncore model to implement a async http server

Usage
=====
>>> EchoServer('0.0.0.0', port=8080, handlers={}, settings={})
>>> asyncore.loop()

"""

import logging
import asyncore
import socket
from karlooper.http_parser.http_parser import HttpParser
from karlooper.config.config import SOCKET_RECEIVE_SIZE, CLIENT_CONNECT_TO_SERVER_NUM
from karlooper.utils import PY3

if PY3:
    unicode = str


class EchoHandler(asyncore.dispatcher_with_send):
    def __init__(self, async_socket, handlers, settings):
        """async echo handler based on asyncore.dispatcher_with_send

        :param async_socket: the socket object
        :param handlers: handlers mapping
        :param settings: settings config

        """
        self.logger = logging.getLogger()
        self.__handlers = handlers
        self.__settings = settings
        asyncore.dispatcher_with_send.__init__(self, sock=async_socket)

    def handle_read(self):
        try:
            if PY3:
                data = self.recv(SOCKET_RECEIVE_SIZE).decode("utf-8")
            else:
                data = self.recv(SOCKET_RECEIVE_SIZE)
            http_parser = HttpParser(data=data, handlers=self.__handlers, settings=self.__settings)
            response_data = http_parser.parse()
            while not (isinstance(response_data, str) or isinstance(response_data, unicode)):
                response_data = http_parser.parse()
            if PY3:
                response_data = response_data.encode('utf-8')
            self.send(response_data)
        except Exception as e:
            self.logger.info("echo handler error: %s" % str(e))


class EchoServer(asyncore.dispatcher):
    def __init__(self, host, port, handlers, settings):
        """async echo server based on asyncore.dispatcher

        :param host: http host
        :param port: http port
        :param handlers: handlers mapping
        :param settings: settings config

        """
        self.__handlers = handlers
        self.__settings = settings
        asyncore.dispatcher.__init__(self)
        self.create_socket(socket.AF_INET, socket.SOCK_STREAM)
        self.set_reuse_addr()
        self.bind((host, port))
        self.listen(CLIENT_CONNECT_TO_SERVER_NUM)

    def handle_accept(self):
        pair = self.accept()
        if pair is not None:
            sock, address = pair
            EchoHandler(sock, self.__handlers, self.__settings)



# -*-coding:utf-8-*-
"""

application
~~~~~~~~~~~

Use this model to initialize web application.

Usage
=====
>>> from karlooper.web import IOModel
>>> from karlooper.web.application import Application
>>> application = Application(handlers={}, settings={}, port=8080, log_conf="./config.log")
>>> application.run(io_model=IOModel.POLL)
server run on port: 8080
run with poll

>>> application = Application(handlers={}, settings={}, log_conf="./config.log")
>>> application.listen(8000)
>>> application.run(io_model=IOModel.POLL)
server run on port: 8000
run with poll

"""

import socket
import select
from karlooper.logger.logger import init_logger
from karlooper.web import IOModel
from karlooper.web.__async_core_server import EchoServer, asyncore
from karlooper.web.http_connection import HttpConnection
from karlooper.web.http_io_buffer import HttpIOBuffer
from karlooper.web.http_io_routine_pool import HttpIORoutinePool
from karlooper.http_parser.http_parser import HttpParser
from karlooper.config import get_cli_data, set_cli_data
from karlooper.config.config import SOCKET_RECEIVE_SIZE, DEFAULT_PORT, CLIENT_CONNECT_TO_SERVER_NUM
from karlooper.utils import PY3

if PY3:
    unicode = str

__author__ = 'karlvorndoenitz@gmail.com'


class Application(object):
    def __init__(self, handlers, settings=None, **kwargs):
        """

        :param handlers: handlers mapping, dict type
        :param settings: settings mapping, dict type
        :param kwargs: options

        """
        self.settings = settings
        self.handlers = handlers
        set_cli_data(self.settings)
        set_cli_data(kwargs)
        cli_data = get_cli_data()
        self.port = int(cli_data.get("port", DEFAULT_PORT))
        log_conf = self.settings.get("log_conf", None) if self.settings else kwargs.get("log_conf", None)
        self.logger = init_logger(config_path=log_conf)
        self.EOL1 = '\n\n'
        self.EOL2 = '\n\r\n'
        self.response = ""

    def listen(self, port):
        """listen port

        :param port: port that application listened
        :return: None

        """
        self.port = int(port)

    def __run_epoll(self):
        """
        run the application use epoll
        """
        server_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        server_socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        server_socket.bind(('0.0.0.0', self.port))
        server_socket.listen(CLIENT_CONNECT_TO_SERVER_NUM)  # the number of client that connect to server
        server_socket.setblocking(0)  # set 0 not block other block
        server_socket.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)
        epoll = select.epoll()
        epoll.register(server_socket.fileno(), select.EPOLLIN)
        try:
            http_connection = HttpConnection()
            http_io_buffer = HttpIOBuffer()
            http_io_routine_pool = HttpIORoutinePool()
            events_buf = []
            while True:
                events = epoll.poll(1) + events_buf
                events_buf = []
                for fileno, event in events:
                    try:
                        if fileno == server_socket.fileno():  # if request come
                            connection, address = server_socket.accept()  # waiting income connection
                            connection.setblocking(0)  # none block
                            epoll.register(connection.fileno(), select.EPOLLIN)  # register socket read event to epoll
                            http_connection.add_connection(connection.fileno(), connection)
                            http_io_buffer.add_request(connection.fileno(), '')
                            http_io_buffer.add_response(connection.fileno(), self.response)
                        elif event & select.EPOLLIN:  # when data in os's read buffer area
                            http_parser = http_io_routine_pool.get(file_no=fileno)
                            if http_parser:
                                data = http_parser.parse()
                                if isinstance(data, str) or isinstance(data, unicode):
                                    http_io_buffer.add_response(
                                        fileno,
                                        http_io_buffer.get_response(fileno) + data
                                    )
                                    epoll.modify(fileno, select.EPOLLOUT)  # change file number to epoll out mode
                                    http_io_routine_pool.remove(fileno)
                                else:  # if coroutine
                                    http_io_routine_pool.add(fileno, http_parser)
                                    events_buf.append((fileno, event))
                            else:
                                if PY3:
                                    http_request_buffer = http_connection.get_connection(fileno).recv(
                                        SOCKET_RECEIVE_SIZE).decode("utf-8")
                                else:
                                    http_request_buffer = http_connection.get_connection(fileno).recv(
                                        SOCKET_RECEIVE_SIZE)
                                http_io_buffer.add_request(
                                    fileno,
                                    http_io_buffer.get_request(fileno) + http_request_buffer
                                )
                                if self.EOL1 in http_io_buffer.get_request(fileno) \
                                        or self.EOL2 in http_io_buffer.get_request(fileno):
                                    request_data = http_io_buffer.get_request(fileno)[:-2] \
                                        if http_io_buffer.get_request(fileno).endswith("\r\n") \
                                        else http_io_buffer.get_request(fileno)
                                    http_parser = HttpParser(
                                        request_data,
                                        self.handlers,
                                        settings=self.settings
                                    )
                                    data = http_parser.parse()
                                    if isinstance(data, str) or isinstance(data, unicode):
                                        http_io_buffer.add_response(
                                            fileno,
                                            http_io_buffer.get_response(fileno) + data
                                        )
                                        epoll.modify(fileno, select.EPOLLOUT)  # change file number to epoll out mode
                                        http_io_routine_pool.remove(fileno)
                                    else:  # if coroutine
                                        http_io_routine_pool.add(fileno, http_parser)
                                        events_buf.append((fileno, event))
                                else:
                                    http_connection.remove_connection(fileno)
                                    http_io_buffer.remove_request(fileno)
                                    http_io_buffer.remove_response(fileno)
                                    http_io_routine_pool.remove(fileno)
                                    epoll.unregister(fileno)
                        elif event & select.EPOLLOUT:  # if out mode
                            http_response_message = http_io_buffer.get_response(fileno)
                            if PY3:
                                http_response_message = http_response_message.encode('utf-8')
                            bytes_written = http_connection.get_connection(fileno).send(
                                http_response_message
                            )
                            http_io_buffer.add_response(fileno, http_io_buffer.get_response(fileno)[bytes_written:])
                            if len(http_io_buffer.get_response(fileno)) == 0:  # if file sent
                                http_connection.get_connection(fileno).shutdown(socket.SHUT_RDWR)
                                epoll.modify(fileno, select.EPOLLHUP)
                        elif event & select.EPOLLHUP:  # if message sent and file number in epoll is hup
                            epoll.unregister(fileno)  # remove file number from epoll
                            http_connection.get_connection(fileno).close()  # close connection
                            http_connection.remove_connection(fileno)  # delete connection from connections dict
                    except Exception as e:
                        self.logger.info("error in __run_epoll: %s", str(e))
                        http_connection.remove_connection(fileno)
                        http_io_buffer.remove_request(fileno)
                        http_io_buffer.remove_response(fileno)
                        http_io_routine_pool.remove(fileno)
                        self.logger.info("fileno is: %s", str(fileno))
                        epoll.close()
                        epoll = select.epoll()
                        epoll.register(server_socket.fileno(), select.EPOLLIN)
        finally:
            epoll.unregister(server_socket.fileno())
            epoll.close()
            server_socket.close()

    def __run_kqueue(self):
        """
        run server use kqueue
        """
        server_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        server_socket.bind(('0.0.0.0', self.port))
        server_socket.listen(CLIENT_CONNECT_TO_SERVER_NUM)
        kq = select.kqueue()
        http_connection = HttpConnection()
        http_io_buffer = HttpIOBuffer()
        http_io_routine_pool = HttpIORoutinePool()
        index = 1
        events = [
            select.kevent(server_socket.fileno(), select.KQ_FILTER_READ, select.KQ_EV_ADD),
            # select.kevent(server_socket.fileno(), select.KQ_FILTER_WRITE, select.KQ_EV_DELETE),
            # select.kevent(server_socket.fileno(), select.KQ_FILTER_READ, select.KQ_EV_ADD),
            # select.kevent(server_socket.fileno(), select.KQ_FILTER_WRITE, select.KQ_EV_DELETE),
        ]
        events_buf = []
        while True:
            try:
                event_list = kq.control(events, 128, 0.0001) + events_buf
                events_buf = []
            except select.error as e:
                self.logger.error("error in __run_kqueue: %s", str(e))
                break
            if event_list:
                for each in event_list:
                    if each.ident == server_socket.fileno():
                        index += 1
                        conn, addr = server_socket.accept()
                        http_connection.add_connection(index, conn)
                        events.append(
                            select.kevent(
                                http_connection.get_connection(index).fileno(),
                                select.KQ_FILTER_READ,
                                select.KQ_EV_ADD,
                                udata=index
                            )
                        )
                    else:
                        try:
                            if each.udata >= 1 and each.filter == select.KQ_FILTER_READ:
                                http_parser = http_io_routine_pool.get(file_no=each.udata)
                                if http_parser:
                                    data = http_parser.parse()
                                    if isinstance(data, str) or isinstance(data, unicode):
                                        http_io_routine_pool.remove(each.udata)
                                        http_io_buffer.add_response(each.udata, data)
                                        events.append(
                                            select.kevent(
                                                http_connection.get_connection(each.udata).fileno(),
                                                select.KQ_FILTER_WRITE,
                                                select.KQ_EV_ADD,
                                                udata=each.udata
                                            )
                                        )
                                        events.remove(select.kevent(
                                            http_connection.get_connection(each.udata).fileno(),
                                            select.KQ_FILTER_READ,
                                            select.KQ_EV_ADD,
                                            udata=each.udata)
                                        )
                                    else:  # if coroutine
                                        http_io_routine_pool.add(each.udata, http_parser)
                                        events_buf.append(each)
                                else:
                                    conn = http_connection.get_connection(each.udata)
                                    if PY3:
                                        request_data = conn.recv(SOCKET_RECEIVE_SIZE).decode("utf-8")
                                    else:
                                        request_data = conn.recv(SOCKET_RECEIVE_SIZE)
                                    request_data = request_data[:-2] if request_data.endswith("\r\n") else request_data
                                    http_parser = HttpParser(
                                        request_data,
                                        handlers=self.handlers,
                                        settings=self.settings
                                    )
                                    data = http_parser.parse()
                                    if isinstance(data, str) or isinstance(data, unicode):
                                        http_io_buffer.add_response(each.udata, data)
                                        events.append(
                                            select.kevent(
                                                http_connection.get_connection(each.udata).fileno(),
                                                select.KQ_FILTER_WRITE,
                                                select.KQ_EV_ADD,
                                                udata=each.udata
                                            )
                                        )
                                        events.remove(select.kevent(
                                            http_connection.get_connection(each.udata).fileno(),
                                            select.KQ_FILTER_READ,
                                            select.KQ_EV_ADD,
                                            udata=each.udata)
                                        )
                                    else:  # if coroutine
                                        http_io_routine_pool.add(each.udata, http_parser)
                                        events_buf.append(each)
                            elif each.udata >= 1 and each.filter == select.KQ_FILTER_WRITE:
                                conn = http_connection.get_connection(each.udata)
                                data = http_io_buffer.get_response(each.udata)
                                if PY3:
                                    data = data.encode('utf-8')
                                conn.send(data)
                                events.remove(select.kevent(
                                    http_connection.get_connection(each.udata).fileno(),
                                    select.KQ_FILTER_WRITE,
                                    select.KQ_EV_ADD,
                                    udata=each.udata)
                                )
                                conn.close()
                                http_connection.remove_connection(each.udata)
                        except Exception as e:
                            self.logger.info("error in __run_kqueue event list: %s", str(e))
                            self.logger.info("each filter: %s", each.filter)
                            self.__remove_event(events, each)
                            http_connection.remove_connection(each.udata)
                            http_io_buffer.remove_request(each.udata)
                            http_io_buffer.remove_response(each.udata)
                            http_io_routine_pool.remove(each.udata)
                            kq.close()
                            kq = select.kqueue()
        server_socket.close()

    def __run_poll(self):
        """
        run server use poll, I will modify __run_poll and __run_epoll in the future
        """
        server_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        server_socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        server_socket.bind(('0.0.0.0', self.port))
        server_socket.listen(CLIENT_CONNECT_TO_SERVER_NUM)  # the number of client that connect to server
        server_socket.setblocking(0)  # set 0 not block other block
        server_socket.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)
        poll = select.poll()
        poll.register(server_socket.fileno(), select.POLLIN)
        try:
            http_connection = HttpConnection()
            http_io_buffer = HttpIOBuffer()
            http_io_routine_pool = HttpIORoutinePool()
            events_buf = []
            while True:
                events = poll.poll(1) + events_buf
                events_buf = []
                for fileno, event in events:
                    try:
                        if fileno == server_socket.fileno():  # if request come
                            connection, address = server_socket.accept()  # waiting income connection
                            connection.setblocking(0)  # none block
                            poll.register(connection.fileno(), select.POLLIN)  # register socket read event to poll
                            http_connection.add_connection(connection.fileno(), connection)
                            http_io_buffer.add_request(connection.fileno(), '')
                            http_io_buffer.add_response(connection.fileno(), self.response)
                        elif event & select.POLLIN:  # when data in os's read buffer area
                            http_parser = http_io_routine_pool.get(file_no=fileno)
                            if http_parser:
                                data = http_parser.parse()
                                if isinstance(data, str) or isinstance(data, unicode):
                                    http_io_buffer.add_response(
                                        fileno,
                                        http_io_buffer.get_response(fileno) + data
                                    )
                                    poll.modify(fileno, select.POLLOUT)  # change file number to epoll out mode
                                    http_io_routine_pool.remove(fileno)
                                else:  # if coroutine
                                    http_io_routine_pool.add(fileno, http_parser)
                                    events_buf.append((fileno, event))
                            else:
                                if PY3:
                                    http_request_buffer = http_connection.get_connection(fileno).recv(
                                        SOCKET_RECEIVE_SIZE).decode("utf-8")
                                else:
                                    http_request_buffer = http_connection.get_connection(fileno).recv(
                                        SOCKET_RECEIVE_SIZE)
                                http_io_buffer.add_request(
                                    fileno,
                                    http_io_buffer.get_request(fileno) + http_request_buffer
                                )
                                if self.EOL1 in http_io_buffer.get_request(fileno) \
                                        or self.EOL2 in http_io_buffer.get_request(fileno):
                                    request_data = http_io_buffer.get_request(fileno)[:-2] \
                                        if http_io_buffer.get_request(fileno).endswith("\r\n") \
                                        else http_io_buffer.get_request(fileno)
                                    http_parser = HttpParser(
                                        request_data,
                                        self.handlers,
                                        settings=self.settings
                                    )
                                    data = http_parser.parse()
                                    if isinstance(data, str) or isinstance(data, unicode):
                                        http_io_buffer.add_response(
                                            fileno,
                                            http_io_buffer.get_response(fileno) + data
                                        )
                                        poll.modify(fileno, select.POLLOUT)  # change file number to epoll out mode
                                        http_io_routine_pool.remove(fileno)
                                    else:  # if coroutine
                                        http_io_routine_pool.add(fileno, http_parser)
                                        events_buf.append((fileno, event))
                                else:
                                    http_connection.remove_connection(fileno)
                                    http_io_buffer.remove_request(fileno)
                                    http_io_buffer.remove_response(fileno)
                                    http_io_routine_pool.remove(fileno)
                                    poll.unregister(fileno)
                        elif event & select.POLLOUT:  # if out mode
                            http_response_message = http_io_buffer.get_response(fileno)
                            if PY3:
                                http_response_message = http_response_message.encode('utf-8')
                            bytes_written = http_connection.get_connection(fileno).send(
                                http_response_message
                            )
                            http_io_buffer.add_response(fileno, http_io_buffer.get_response(fileno)[bytes_written:])
                            if len(http_io_buffer.get_response(fileno)) == 0:  # if file sent
                                http_connection.get_connection(fileno).shutdown(socket.SHUT_RDWR)
                                poll.modify(fileno, select.POLLHUP)
                        elif event & select.POLLHUP:  # if message sent and file number in poll is hup
                            poll.unregister(fileno)  # remove file number from poll
                            http_connection.get_connection(fileno).close()  # close connection
                            http_connection.remove_connection(fileno)  # delete connection from connections dict
                    except Exception as e:
                        self.logger.info("error in __run_poll: %s", str(e))
                        http_connection.remove_connection(fileno)
                        http_io_buffer.remove_request(fileno)
                        http_io_buffer.remove_response(fileno)
                        http_io_routine_pool.remove(fileno)
                        self.logger.info("fileno is: %s", str(fileno))
                        poll.unregister(fileno)
        finally:
            poll.unregister(server_socket.fileno())
            poll.close()
            server_socket.close()

    def __run_async_io(self):
        """
        run server use asyncore
        """
        EchoServer('0.0.0.0', self.port, self.handlers, self.settings)
        asyncore.loop()

    def __remove_event(self, events, each):
        """remove event from events

        :param events: the list contain some events
        :param each: the event will be removed
        :return: None

        """
        self.logger.warning("remove event with udata: %s", str(each.udata))
        for event in events:
            if event.ident == each.ident:
                events.remove(event)
                break

    def run(self, io_model=None):
        """run the web server

        :param io_model: os io model, EPOLL 0 KQUEUE 1 POLL 2
        :return: None

        """
        print("server run on port: %d" % self.port)
        self.logger.info("server run on port: %d" % self.port)
        if io_model:
            if io_model == IOModel.EPOLL and hasattr(select, "epoll"):
                print("run with epoll")
                self.logger.info("run with epoll")
                self.__run_epoll()
            elif io_model == IOModel.KQUEUE and hasattr(select, "kqueue"):
                print("run with kqueue")
                self.logger.info("run with kqueue")
                self.__run_kqueue()
            elif io_model == IOModel.POLL and hasattr(select, "poll"):
                print("run with poll")
                self.logger.info("run with poll")
                self.__run_poll()
            elif io_model == IOModel.ASYNCIO:
                print("run with asyncore")
                self.logger.info("run with asyncore")
                self.__run_async_io()
        else:
            if hasattr(select, "epoll"):
                print("run with epoll")
                self.logger.info("run with epoll")
                self.__run_epoll()
            elif hasattr(select, "kqueue"):
                print("run with kqueue")
                self.logger.info("run with kqueue")
                self.__run_kqueue()
            elif hasattr(select, "poll"):
                print("run with poll")
                self.logger.info("run with poll")
                self.__run_poll()
            else:
                print("run with asyncore")
                self.logger.info("run with asyncore")
                self.__run_async_io()
        print("server start failed!")
        self.logger.info("server start failed!")



# -*-coding:utf-8-*-
"""

security
~~~~~~~~

Use this model to encrypt string.

Usage
=====
>>> d = Encryption()
>>> d.input_key("123456789")
>>> s = "/static/hello.js"
>>> a = d.encode(s)
>>> print a
b14f1453ceddc91e492fbe883d552a2e
>>> b = d.decode(a)
>>> print b
/static/hello.js

"""

import base64
from karlooper.utils import PY3


class Encryption(object):

    def __init__(self):
        self.__key = ""

    def input_key(self, key):
        """ set base key

        :param key: str type, the base key
        :return: None

        """
        self.__key = key

    def encode(self, s):
        """ encode a string

        :param s: the string will be encoded
        :return: encoded result

        """
        ss = s + self.__key
        if PY3:
            return base64.b64encode(ss.encode()).decode()
        else:
            return base64.b64encode(ss)

    def decode(self, s):
        """ decode a string

        :param s: the string will be decoded
        :return: decoded result

        """
        missing_padding = 4 - len(s) % 4
        if missing_padding:
            s += '=' * missing_padding
        if PY3:
            decode_result = base64.b64decode(s.encode()).decode()
        else:
            decode_result = base64.b64decode(s)
        result = decode_result[:len(decode_result)-len(self.__key)]
        return result



# -*-encoding:utf-8-*-

import functools


class Future(object):
    pass


class CoroutineException(Exception):
    message = None


class Coroutine(object):
    coroutine = None
    context = None


class CoroutinePool(object):
    cid_list = [0]
    pool = {}

    @property
    def size(self):
        """ get coroutine pool's size

        :return: size

        """
        return len(self.cid_list)

    def add(self, cid, coroutine_obj):
        """ add a coroutine object to coroutine pool

        :param cid: coroutine id
        :param coroutine_obj: coroutine object
        :return: None

        """
        self.cid_list.append(cid)
        self.pool[cid] = coroutine_obj

    def remove(self, cid):
        """ remove the coroutine object by id

        :param cid: coroutine's id
        :return: None

        """
        self.cid_list.remove(cid)
        del self.pool[cid]


__coroutine_pool = CoroutinePool()


def koroutine(method):

    @functools.wraps(method)
    def _deco(*args, **kwargs):
        global __coroutine_pool
        if not hasattr(method, "coroutine_id"):
            cid = __coroutine_pool.cid_list[-1] + 1
            setattr(method, "coroutine_id", cid)
            coroutine_obj = Coroutine()
            generator = method(*args, **kwargs)
            coroutine_obj.coroutine = generator
            coroutine_obj.context = next(generator)
            __coroutine_pool.add(cid, coroutine_obj)
            return Future
        cid = getattr(method, "coroutine_id")
        coroutine_obj = __coroutine_pool.pool.get(cid, None)
        if not coroutine_obj:
            ce = CoroutineException()
            ce.message = "Coroutine Error"
            raise ce
        coroutine = getattr(coroutine_obj, "coroutine")
        result = next(coroutine, Future)
        if not isinstance(result, type):
            setattr(coroutine_obj, "context", result)
            return Future
        result = getattr(coroutine_obj, "context")
        del method.coroutine_id
        __coroutine_pool.remove(cid)
        return result

    return _deco



# -*-coding:utf-8-*-
"""

response
~~~~~~~~

Use this model to customize your http response web page.

Usage
=====
>>> from karlooper.web.response import HTTPResponse404
>>> def not_found(self):
...     return "<h1>Page Not Found</h1>"
...
>>> HTTPResponse404.data = not_found

"""
from karlooper.config.config import HttpStatus, HttpStatusMsg
from karlooper.template import render


class HTTPResponse(object):
    status = 0
    message = ""
    data = ""
    
    def __init__(self, **kwargs):
        self.__settings = kwargs
        
    def render(self, template_path, **kwargs):
        """ customize your HTTPResponse web page use template

        :param template_path: template path
        :param kwargs: parameters
        :return: html formatted string

        """
        static_path = self.__settings.get("template", ".")
        template_path = static_path + template_path
        return render(template_path, **kwargs)


class HTTPResponse404(HTTPResponse):
    status = HttpStatus.NOT_FOUND
    message = HttpStatusMsg.NOT_FOUND

    def data(self):
        return str(self.status)


class HTTPResponse405(HTTPResponse):
    status = HttpStatus.METHOD_NOT_ALLOWED
    message = HttpStatusMsg.METHOD_NOT_ALLOWED

    def data(self):
        return str(self.status)


class HTTPResponse500(HTTPResponse):
    status = HttpStatus.SERVER_ERROR
    message = HttpStatusMsg.SERVER_ERROR

    def data(self):
        return ""



# -*-coding:utf-8-*-
"""

security
~~~~~~~~

Use this model to encrypt string.

Usage
=====
>>> d = StrEncryption()
>>> d.input_key("123456789")
>>> s = "/static/hello.js"
>>> a = d.encode(s)
>>> print a
L3N0YXRpYy9oZWxsby5qczEyMzQ1Njc4OQ==
>>> b = d.decode(a)
>>> print b
/static/hello.js

"""
from karlooper.utils import PY3

__author__ = 'karlvorndoenitz@gmail.com'


try:
    if PY3:
        from karlooper.utils.encryption import encrypt_str, decrypt_str
    else:
        from encryption import encrypt_str, decrypt_str


    class StrEncryption(object):
        def __init__(self):
            self.__key = ""

        def input_key(self, key):
            """ set input key

            :param key: input key
            :return: None

            """
            self.__key = key

        def encode(self, s):
            """ encode string

            :param s: the string will be encoded
            :return: the string be encoded

            """
            return encrypt_str(s + self.__key)

        def decode(self, s):
            """ decode string

            :param s: the string will be decoded
            :return: the string be decoded

            """
            decode_result = decrypt_str(s)
            return decode_result[:len(decode_result)-len(self.__key)]

except ImportError:
    try:
        if PY3:
            from karlooper.utils.base64encrypt import Encryption as StrEncryption
        else:
            from base64encrypt import Encryption as StrEncryption
    except ImportError:
        if PY3:
            from karlooper.utils.des_encrypt import DES as StrEncryption
        else:
            from des_encrypt import DES as StrEncryption



# -*-coding:utf-8-*-

import datetime
import logging
from karlooper.router.router import Router
from karlooper.config.config import ContentType, RESPONSE_HEAD_MESSAGE
from karlooper.web.response import HTTPResponse404, HTTPResponse500

__author__ = 'karlvorndoenitz@gmail.com'


class HttpParser(object):
    def __init__(self, data, handlers, settings=None):
        """

        :param data: the request data from socket
        :param settings: url settings

        """
        self.logger = logging.getLogger()
        self.response_header = RESPONSE_HEAD_MESSAGE
        self.data = data
        self.settings = settings if settings else {}
        self.handlers = handlers
        self.host = self.settings.get("host", "Karlooper")

    def parse(self):
        return self.__parse_data()

    def __parse_data(self):
        """

        parse http request data to dict and get handler's response data

        :return: http response data

        """
        http_message = self.data.split("\r\n\r\n")
        http_request_head = http_message[0]
        http_request_line = http_request_head.split("\r\n")[0].split(" ")
        http_method = http_request_line[0].lower()
        http_url = http_request_line[1]
        http_version = http_request_line[2]
        http_head_list = http_request_head.split("\r\n")[1:]
        if "" in http_head_list:
            http_head_list.remove("")
        header = dict((header.split(": ")[0].lower(), header.split(": ")[1]) for header in http_head_list)
        http_body = http_message[1] if len(http_message) > 1 else None
        http_message_dict = dict({})
        http_message_dict["url"] = http_url
        http_message_dict["method"] = http_method
        http_message_dict["version"] = http_version
        http_message_dict["header"] = header
        http_message_dict["body"] = http_body
        now = datetime.datetime.now()
        now_time = now.strftime("%a, %d %b %Y %H:%M:%S GMT")
        status = dict({})
        status["date"] = now_time
        status["host"] = self.host
        url = http_url.split("?")[0]
        handler = Router(self.handlers, url).get_handler()
        path_param = handler[1]
        handler = handler[0]
        if not handler:
            data = HTTPResponse404(**self.settings).data()
            status["status"] = HTTPResponse404.status
            status["content_type"] = ContentType.HTML
            status["content_length"] = len(data)
            status["status_msg"] = HTTPResponse404.message
        else:
            try:
                handler_init = handler(http_message_dict, self.data, self.settings)
                if path_param.status:
                    handler_init.set_path_param(path_param.value)
                handler_init.before_request()
                pro_function = getattr(handler_init, http_method)
                data = pro_function()
                if isinstance(data, type):
                    return data
                try:
                    handler_init_k = data[-1]
                    http_response_header = handler_init_k.get_response_header()
                except Exception as e:
                    self.logger.info("static handler response: %s", str(e))
                    http_response_header = handler_init.get_response_header()
                self.response_header = self.response_header.replace("\r\n\r\n", http_response_header)
                status["status"] = data[1]
                status["status_msg"] = data[2]
                data = data[0]
                handler_init.teardown_request()
            except Exception as e:
                status["status"] = HTTPResponse500.status
                status["status_msg"] = HTTPResponse500.message
                data = HTTPResponse500(**self.settings).data() or str(e)
            status["content_length"] = len(data)
        response_data = self.response_header % status + data
        return response_data



# -*-encoding:utf-8-*-


class HttpIORoutinePool(object):
    """

    This class contains four methods.

    def __init__(self): initialize

    def add(self, file_no, coroutine): add co-routine

    def get(self, file_no): get co-routine

    def remove(self, file_no): remove co-routine

    """
    def __init__(self):
        self.pool = {}

    def add(self, file_no, coroutine):
        """add co-routine to pool

        :param file_no: number of file description
        :param coroutine: co-routine
        :return: None

        """
        self.pool[file_no] = coroutine

    def get(self, file_no):
        """ get co-routine from pool

        :param file_no: number of file description
        :return: co-routine

        """
        return self.pool.get(file_no, None)

    def remove(self, file_no):
        """remove co-routine from pool

        :param file_no: number of file description
        :return: None

        """
        if file_no in self.pool:
            del self.pool[file_no]



# -*-coding:utf-8-*-
"""

config
~~~~~~
store global config data

Usage
=====
>>> set_cli_data({"port": 8080})
>>> data = get_cli_data()

"""


__CLI_DATA = dict({})
__GLOBAL_CONF_DATA = dict({})


def set_cli_data(data):
    """ set global config data

    :param data: a dict contain global config data
    :return: None

    """
    global __CLI_DATA
    if not data:
        return
    for key in data:
        __CLI_DATA[key] = data[key]


def get_cli_data():
    """ get data from command line client

    :return: None

    """
    global __CLI_DATA
    return __CLI_DATA


def set_global_conf(key, data):
    """ set global config data
    
    :param key: global data's key
    :param data: global data
    :return: None
    
    """
    global __GLOBAL_CONF_DATA
    if not data:
        return
    __GLOBAL_CONF_DATA[key] = data


def get_global_conf(key):
    """ get data from global config
    
    :param key: the data's key
    :return: the data
    
    """
    global __GLOBAL_CONF_DATA
    return __GLOBAL_CONF_DATA[key]




# -*-coding:utf-8-*-
"""

auto reload
~~~~~~~~~~~

Use this model to watch your project and reload your project when the code was modified.

Usage
=====
>>> from karlooper.autoreload import AutoReload
>>> from karlooper.web.application import Application
>>> application = Application()
>>> auto_reload = AutoReload(application)
>>> auto_reload.run()

"""

import time


class AutoReload(object):
    def __init__(self, **kwargs):
        pass

    def __check(self):
        pass

    def __application(self):
        pass

    def run(self):
        thread_obj = self.__application()
        while True:
            time.sleep(1)
            if self.__check():
                thread_obj.stop()
                thread_obj = self.__application()



#!/usr/bin/env python
import os
import platform
from setuptools import setup, Extension
from karlooper import __version__

# load requirements file
requirements = 'requirements.txt'
if os.path.exists(requirements):
    with open(requirements) as f:
        required = f.read().splitlines()
else:
    with open('karlooper.egg-info/requires.txt') as f:
        required = f.read().splitlines()

# whether need add c extension
kwargs = {}
if platform.python_implementation() == 'CPython' and platform.system().lower() not in ['windows']:
    kwargs['ext_modules'] = [
        Extension(
            'karlooper.utils.encryption',
            sources=['karlooper/utils/encryption.c'],
            extra_compile_args=["-Wno-char-subscripts"]
        ),
    ]

setup(
    name='karlooper',
    version=__version__,
    install_requires=required,
    packages=[
        'karlooper',
        'karlooper.autoreload',
        'karlooper.config',
        'karlooper.coroutine',
        'karlooper.escape',
        'karlooper.http_parser',
        'karlooper.logger',
        'karlooper.router',
        'karlooper.template',
        'karlooper.utils',
        'karlooper.web'
    ],
    package_data={
        '': ['*.txt', '*.pyi', '*.conf']
    },
    url='https://github.com/karldoenitz/karlooper',
    license='MIT',
    author='lizhihao',
    author_email='karlvorndoenitz@gmail.com',
    description='karloop plus micro-framework',
    long_description=open("README.rst").read(),
    classifiers=[
        "Environment :: Web Environment",
        "Intended Audience :: Developers",
        "Operating System :: OS Independent",
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7'
    ],
    **kwargs
)



# -*-coding:utf-8-*-

import os
import logging
import logging.config
from karlooper.config import get_cli_data


def init_logger(config_path=None):
    """get the logger object

    :param config_path: log config file
    :return: logger

    """
    cli_data = get_cli_data()
    log_enable = cli_data.get("log_enable")
    if not log_enable:
        return logging
    if isinstance(log_enable, str) and log_enable.lower() != "true":
        return logging
    path = os.path.dirname(__file__) + "/log.conf"
    config_file = config_path or path
    logging.config.fileConfig(config_file)
    logger = logging.getLogger()
    return logger
